/////////////////////////////////////////////////////////////////////////////
// Project:     wxIView
// Purpose:     Complex wxWidgets sample
// Name:        wxoptionsdialog.cpp
// Author:      Daniel Nell
// Created:     2006/02/16
// Copyright:   (c) Daniel Nell
// Licence:     wxWindows licence
// Modified by:
/////////////////////////////////////////////////////////////////////////////

// For compilers that support precompilation, includes "wx/wx.h".
#include <wx/wxprec.h>
#ifndef WX_PRECOMP
   #include <wx/wx.h>
#endif

// Include additional wxWidgets headers.
#include <wx/config.h>
#include <wx/statline.h>
#include <wx/listbook.h>

// Include private headers.
#include "wxoptionsdialog.h"


//////////////////////////////////////////////////////////
// class wxOptionStaticLine
//
wxOptionStaticLine::wxOptionStaticLine()
{
}

void wxOptionStaticLine::Setup( wxWindow *poWindow, int nId )
{
    m_poStaticLine= new wxStaticLine( poWindow, nId );
    if( !m_poStaticLine )
    {
        wxLogError( _( "Cannot create static line pseudo option." ) );
        return;
    }
}

bool wxOptionStaticLine::AddToSizer( wxBoxSizer *poSizer )
{
    return m_poStaticLine
           ? (bool)poSizer->Add( m_poStaticLine, 0, wxGROW | wxLEFT | wxRIGHT, 1 )
           : false;
}


//////////////////////////////////////////////////////////
// class wxOptionCheckBox
//
wxOptionCheckBox::wxOptionCheckBox( const wxString &strText, bool *pbValue )
 : wxOptionBase()
{
    m_pbValue= pbValue;
    m_strText= strText;
}

bool wxOptionCheckBox::AddToSizer( wxBoxSizer *poSizer )
{
    return m_poCheckBox
           ? (bool)poSizer->Add( m_poCheckBox, 0, wxALIGN_LEFT | wxALL, 1 )
           : false;
}

void wxOptionCheckBox::Setup( wxWindow *poWindow, int nId )
{
    m_poCheckBox= new wxCheckBox( poWindow, nId, m_strText );
    if( !m_poCheckBox )
    {
        wxLogError( _( "Cannot create option checkbox." ) );
        return;
    }
    m_poCheckBox->SetValue( *m_pbValue );
}

void wxOptionCheckBox::StoreValue()
{
    (*m_pbValue)= m_poCheckBox->GetValue();
}


//////////////////////////////////////////////////////////
// class wxOptionTextBox
//
wxOptionTextBox::wxOptionTextBox( const wxString &strText, wxString *pstrValue, int nWidth )
{
    m_strText= strText;
    m_pstrValue= pstrValue;
    m_nWidth= nWidth;
}

bool wxOptionTextBox::AddToSizer( wxBoxSizer *poSizer )
{
    if( !m_poTextCtrl )
    {
        return false;
    }

    wxBoxSizer *poHSizer= new wxBoxSizer( wxHORIZONTAL );

    if( !poHSizer )
    {
        return false;
    }
    if(    !poHSizer->Add( new wxStaticText( m_poWindow, wxID_ANY, m_strText ), 0,
                           wxALIGN_CENTER_VERTICAL | wxALIGN_LEFT )
        || !poHSizer->Add( m_poTextCtrl, 0,
                           wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT )
    )
    {
        delete poHSizer;
        return false;
    }
    return (bool)poSizer->Add( poHSizer, 0, wxALIGN_RIGHT | wxALL, 1 );
}

void wxOptionTextBox::Setup( wxWindow *poWindow, int nId )
{
    m_poWindow= poWindow;

    m_poTextCtrl= new wxTextCtrl( m_poWindow, nId, *m_pstrValue, wxDefaultPosition, wxSize( m_nWidth, -1 ) );
    if( !m_poTextCtrl )
    {
        wxLogError( _( "Cannot create option textbox." ) );
        return;
    }
}

void wxOptionTextBox::StoreValue()
{
    (*m_pstrValue)= m_poTextCtrl->GetValue();
}



//////////////////////////////////////////////////////////
// class wxOptionTextBox
//
wxOptionComboBox::wxOptionComboBox( const wxString &strText, wxString *pstrValue, int nWidth )
{
    m_strText= strText;
    m_pstrValue= pstrValue;
    m_nWidth= nWidth;
}

wxOptionComboBox::~wxOptionComboBox()
{
    // Get the global configuration object and set path.
    wxConfigBase *poConfig = wxConfigBase::Get();
    poConfig->SetPath( wxString::Format( wxT( "/wxOptionComboBox/%d" ), m_nId ) );

    // Write history to config (file/registry).
    int nCount= m_astrValueHistory.GetCount();

    for( int i= 0; i < nCount; i++ )
    {
        poConfig->Write( wxString::Format( wxT( "Text%d" ), i ), m_astrValueHistory[i] );
    }
}

bool wxOptionComboBox::AddToSizer( wxBoxSizer *poSizer )
{
    if( !m_poComboBox )
    {
        return false;
    }

    wxBoxSizer *poHSizer= new wxBoxSizer( wxHORIZONTAL );

    if( !poHSizer )
    {
        return false;
    }
    if(    !poHSizer->Add( new wxStaticText( m_poWindow, wxID_ANY, m_strText ), 0,
                           wxALIGN_CENTER_VERTICAL | wxALIGN_LEFT )
        || !poHSizer->Add( m_poComboBox, 0,
                           wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT )
    )
    {
        delete poHSizer;
        return false;
    }
    return (bool)poSizer->Add( poHSizer, 0, wxALIGN_RIGHT | wxALL, 1 );
}

void wxOptionComboBox::Setup( wxWindow *poWindow, int nId )
{
    // ...
    m_poWindow= poWindow;
    m_nId= nId;

    // Get the global configuration object and set path.
    wxConfigBase *poConfig = wxConfigBase::Get();
    poConfig->SetPath( wxString::Format( wxT( "/wxOptionComboBox/%d" ), m_nId ) );

    // ...
    wxString strTextId;
    int i= 0;

    for( strTextId.Printf( wxT( "Text%d" ), i );
         poConfig->HasEntry( strTextId );
         i++, strTextId.Printf( wxT( "Text%d" ), i ) )
    {
        wxString strTemp= poConfig->Read( strTextId, wxT( "" ) );
        m_astrValueHistory.Add( strTemp );
    }

    // Create the combobox.
    m_poComboBox= new wxComboBox( m_poWindow, m_nId, *m_pstrValue, wxDefaultPosition, wxSize( m_nWidth, -1 ),
                                  m_astrValueHistory, wxCB_DROPDOWN | wxCB_SORT );
    if( !m_poComboBox )
    {
        wxLogError( _( "Cannot create option combobox." ) );
        return;
    }
}

void wxOptionComboBox::StoreValue()
{
    // Get the text from the edit control.
    (*m_pstrValue)= m_poComboBox->GetValue();

    // Check if the text is already in the history list and add it if not.
    int nCount= m_astrValueHistory.GetCount();
    int i= 0;

    for( ; i < nCount; i++ )
    {
        if( m_astrValueHistory[i] == (*m_pstrValue) )
        {
            return;
        }
    }
    m_astrValueHistory.Add( (*m_pstrValue) );
}


//////////////////////////////////////////////////////////
// The standard options dialog
//
BEGIN_EVENT_TABLE( wxOptionsDialog, wxDialog )
    EVT_BUTTON( wxID_OK, wxOptionsDialog::OnOk )
END_EVENT_TABLE()


wxOptionsDialog::wxOptionsDialog( wxWindow *poParentWindow )
  : wxDialog( poParentWindow, -1,
              wxTheApp->GetAppName()  + wxT( " - " ) + _( "Options" ),
              wxDefaultPosition,
              wxDefaultSize,
              wxDEFAULT_DIALOG_STYLE /* | wxRESIZE_BORDER */ ),
    m_nOptionCount( 0 ),
    m_nOptionId( WXOPTIONSDIALOGIDOPTION )
{
    // Initialize members. (Nothing to do, yet.)
}

wxOptionsDialog::~wxOptionsDialog()
{
    // Clean up options.
    while( m_nOptionCount > 0 )
    {
        m_nOptionCount--;
        delete m_apoOptions[m_nOptionCount];
    }
}

void wxOptionsDialog::SetupDialog()
{
    // Create layout.
    wxBoxSizer *poSizer= new wxBoxSizer( wxVERTICAL );

    // Create controls.
    for( int i= 0; i < m_nOptionCount; i++ )
    {
        m_apoOptions[i]->AddToSizer( poSizer );
    }
    poSizer->Add( new wxStaticLine( this ), 0, wxGROW | wxLEFT | wxRIGHT, 1 );
    poSizer->Add( CreateButtonSizer( wxOK | wxCANCEL ), 0, wxALIGN_CENTER | wxALL, 5 );

    // Activate.
//    SetSizer( poSizer );
GetParent()->SetSizer( poSizer );
//GetParent()->Layout();
//poSizer->SetSizeHints( GetParent() );
//poSizer->Fit( GetParent() );
//GetParent()->SetAutoLayout( true );
//wxBoxSizer *poMSizer= new wxBoxSizer( wxVERTICAL );
//SetSizer( poMSizer );
//Layout();
    SetAutoLayout( true );
    poSizer->SetSizeHints( this );
    poSizer->Fit( this );
    Centre( wxBOTH | wxCENTRE_ON_SCREEN );
}

bool wxOptionsDialog::AppendOption( wxOptionBase *poOption )
{
    if( !poOption )
    {
        return false;
    }
    if( m_nOptionCount >= WXOPTIONSDIALOGMAXOPTION )
    {
        return false;
    }
    poOption->Setup( GetParent(), m_nOptionId );
    m_nOptionId++;
    m_apoOptions[m_nOptionCount]= poOption;
    m_nOptionCount++;
    return true;
}

void wxOptionsDialog::OnOk( wxCommandEvent& WXUNUSED( oEvent ) )
{
    // Get values.
    for( int i= 0; i < m_nOptionCount; i++ )
    {
        m_apoOptions[i]->StoreValue();
    }

    // End modal dialog mode and set the return value to 'OK'.
    EndModal( wxID_OK );
}


#if __NOTYETRELEASED__ >= 2
//////////////////////////////////////////////////////////
// The list-/choice-/notebook options dialog
//
wxOptionsBookDialog::wxOptionsBookDialog( wxWindow *poParentWindow, const wxString &strPageName,
                                          wxBookCtrlBase *poBookCtrl )
 : wxOptionsDialog( poParentWindow )
{
    // Initialize members.
    m_poBookCtrl= poBookCtrl ? poBookCtrl : new wxListbook( this, wxID_ANY, wxDefaultPosition, wxDefaultSize, wxLB_LEFT );

    // Do further initialization.
    //
    // Add the first page to the ...book.
    AddPage( strPageName );
}

bool wxOptionsBookDialog::AddPage( const wxString &strPageName )
{
    wxPanel *poPanel= new wxPanel( m_poBookCtrl );

    if( !poPanel )
    {
        wxLogError( _( "Cannot create new panel for options to place onto." ) );
        return false;
    }
    if( !m_poBookCtrl->AddPage( poPanel, strPageName, false, -1 ) )
    {
        wxLogWarning( _( "Cannot add new page." ) );
        delete poPanel;
        return false;
    }
    m_poParentForControls= poPanel;
    return true;
}
#endif // __NOTYETRELEASED__
